# btrdb.stream
# Module for Stream and related classes
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: stream.py [] allen@pingthings.io $

"""
Module for Stream and related classes
"""

##########################################################################
## Imports
##########################################################################

import re
import json
import uuid as uuidlib
from copy import deepcopy
from collections.abc import Sequence

from btrdb.utils.buffer import PointBuffer
from btrdb.point import RawPoint, StatPoint
from btrdb.transformers import StreamSetTransformer
from btrdb.exceptions import BTrDBError, InvalidOperation
from btrdb.utils.timez import currently_as_ns, to_nanoseconds
from btrdb.utils.conversion import AnnotationEncoder, AnnotationDecoder
from btrdb.utils.general import pointwidth as pw


##########################################################################
## Module Variables
##########################################################################

INSERT_BATCH_SIZE = 50000
MINIMUM_TIME = -(16 << 56)
MAXIMUM_TIME = (48 << 56) - 1

try:
    RE_PATTERN = re._pattern_type
except Exception:
    RE_PATTERN = re.Pattern


##########################################################################
## Stream Classes
##########################################################################

class Stream(object):
    """
    An object that represents a specific time series stream in the BTrDB database.

    Parameters
    ----------
        btrdb : BTrDB
            A reference to the BTrDB object connecting this stream back to the physical server.
        uuid : UUID
            The unique UUID identifier for this stream.
        db_values : kwargs
            Framework only initialization arguments.  Not for developer use.

    """

    def __init__(self, btrdb, uuid, **db_values):
        db_args = ('known_to_exist', 'collection', 'tags', 'annotations', 'property_version')
        for key in db_args:
            value = db_values.pop(key, None)
            setattr(self, "_{}".format(key), value)
        if db_values:
            bad_keys = ", ".join(db_values.keys())
            raise TypeError("got unexpected db_values argument(s) '{}'".format(bad_keys))

        self._btrdb = btrdb
        self._uuid = uuid

    def refresh_metadata(self):
        """
        Refreshes the locally cached meta data for a stream

        Queries the BTrDB server for all stream metadata including collection,
        annotation, and tags. This method requires a round trip to the server.
        """

        ep = self._btrdb.ep
        self._collection, self._property_version, self._tags, self._annotations, _ = ep.streamInfo(self._uuid, False, True)
        self._known_to_exist = True

        # deserialize annoation values
        self._annotations = {
            key: json.loads(val, cls=AnnotationDecoder)
            for key, val in self._annotations.items()
        }

    def exists(self):
        """
        Check if stream exists

        Exists returns true if the stream exists. This is essential after using
        StreamFromUUID as the stream may not exist, causing a 404 error on
        later stream operations. Any operation that returns a stream from
        collection and tags will have ensured the stream exists already.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            Indicates whether stream is extant in the BTrDB server.
        """

        if self._known_to_exist:
            return True

        try:
            self.refresh_metadata()
            return True
        except BTrDBError as bte:
            if bte.code == 404:
                return False
            raise bte

    def count(self, start=MINIMUM_TIME, end=MAXIMUM_TIME, pointwidth=62, precise=False, version=0):
        """
        Compute the total number of points in the stream

        Counts the number of points in the specified window and version. By
        default returns the latest total count of points in the stream. This
        helper method sums the counts of all StatPoints returned by
        ``aligned_windows``. Because of this, note that the start and end
        timestamps may be adjusted if they are not powers of 2. For smaller
        windows of time, you may also need to adjust the pointwidth to ensure
        that the count granularity is captured appropriately.

        Alternatively you can set the precise argument to True which will
        give an exact count to the nanosecond but may be slower to execute.

        Parameters
        ----------
        start : int or datetime like object, default: MINIMUM_TIME
            The start time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)

        end : int or datetime like object, default: MAXIMUM_TIME
            The end time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)

        pointwidth : int, default: 62
            Specify the number of ns between data points (2**pointwidth).  If the value
            is too large for the time window than the next smallest, appropriate
            pointwidth will be used.

        precise : bool, default: False
            Forces the use of a windows query instead of aligned_windows
            to determine exact count down to the nanosecond.  This will
            be some amount slower than the aligned_windows version.

        version : int, default: 0
            Version of the stream to query

        Returns
        -------
        int
            The total number of points in the stream for the specified window.
        """

        if not precise:
            pointwidth = min(pointwidth, pw.from_nanoseconds(to_nanoseconds(end) - to_nanoseconds(start))-1)
            points = self.aligned_windows(start, end, pointwidth, version)
            return  sum([point.count for point, _ in points])

        depth = 0
        width = to_nanoseconds(end) - to_nanoseconds(start)
        points = self.windows(start, end, width, depth, version)
        return  sum([point.count for point, _ in points])


    @property
    def btrdb(self):
        """
        Returns the stream's BTrDB object.

        Parameters
        ----------
        None

        Returns
        -------
        BTrDB
            The BTrDB database object.

        """
        return self._btrdb

    @property
    def uuid(self):
        """
        Returns the stream's UUID. The stream may nor may not exist
        yet, depending on how the stream object was obtained.

        Returns
        -------
        UUID
            The unique identifier of the stream.


        See Also
        --------
        stream.exists()

        """
        return self._uuid

    @property
    def name(self):
        """
        Returns the stream's name which is parsed from the stream tags.  This
        may require a round trip to the server depending on how the stream was
        acquired.

        Returns
        -------
        str
            The name of the stream.

        """
        return self.tags()["name"]

    @property
    def unit(self):
        """
        Returns the stream's unit which is parsed from the stream tags.  This
        may require a round trip to the server depending on how the stream was
        acquired.

        Returns
        -------
        str
            The unit for values of the stream.

        """
        return self.tags()["unit"]

    @property
    def collection(self):
        """
        Returns the collection of the stream. It may require a round trip to the
        server depending on how the stream was acquired.

        Parameters
        ----------
        None

        Returns
        -------
        str
            the collection of the stream

        """
        if self._collection is not None:
            return self._collection

        self.refresh_metadata()
        return self._collection

    def earliest(self, version=0):
        """
        Returns the first point of data in the stream. Returns None if error
        encountered during lookup or no values in stream.

        Parameters
        ----------
        version : int, default: 0
            Specify the version of the stream to query; if zero, queries the latest
            stream state rather than pinning to a version.

        Returns
        -------
        tuple
            The first data point in the stream and the version of the stream
            the value was retrieved at (tuple(RawPoint, int)).

        """
        start = MINIMUM_TIME
        return self.nearest(start, version=version, backward=False)

    def latest(self, version=0):
        """
        Returns last point of data in the stream. Returns None if error
        encountered during lookup or no values in stream.

        Parameters
        ----------
        version : int, default: 0
            Specify the version of the stream to query; if zero, queries the latest
            stream state rather than pinning to a version.

        Returns
        -------
        tuple
            The last data point in the stream and the version of the stream
            the value was retrieved at (tuple(RawPoint, int)).

        """
        start = MAXIMUM_TIME
        return self.nearest(start, version=version, backward=True)

    def current(self, version=0):
        """
        Returns the point that is closest to the current timestamp, e.g. the latest
        point in the stream up until now. Note that no future values will be returned.
        Returns None if errors during lookup or there are no values before now.

        Parameters
        ----------
        version : int, default: 0
            Specify the version of the stream to query; if zero, queries the latest
            stream state rather than pinning to a version.

        Returns
        -------
        tuple
            The last data point in the stream up until now and the version of the stream
            the value was retrieved at (tuple(RawPoint, int)).
        """
        start = currently_as_ns()
        return self.nearest(start, version, backward=True)

    def tags(self, refresh=False):
        """
        Returns the stream's tags.

        Tags returns the tags of the stream. It may require a round trip to the
        server depending on how the stream was acquired.

        Parameters
        ----------
        refresh: bool
            Indicates whether a round trip to the server should be implemented
            regardless of whether there is a local copy.

        Returns
        -------
        dict
            A dictionary containing the tags.

        """
        if refresh or self._tags is None:
            self.refresh_metadata()

        return deepcopy(self._tags)

    def annotations(self, refresh=False):
        """
        Returns a stream's annotations

        Annotations returns the annotations of the stream (and the annotation
        version). It will always require a round trip to the server. If you are
        ok with stale data and want a higher performance version, use
        Stream.CachedAnnotations().

        Do not modify the resulting map.

        Parameters
        ----------
        refresh: bool
            Indicates whether a round trip to the server should be implemented
            regardless of whether there is a local copy.

        Returns
        -------
        tuple
            A tuple containing a dictionary of annotations and an integer representing
            the version of the metadata (tuple(dict, int)).

        """
        if refresh or self._annotations is None:
            self.refresh_metadata()

        return deepcopy(self._annotations), deepcopy(self._property_version)

    def version(self):
        """
        Returns the current data version of the stream.

        Version returns the current data version of the stream. This is not
        cached, it queries each time. Take care that you do not intorduce races
        in your code by assuming this function will always return the same vaue

        Parameters
        ----------
        None

        Returns
        -------
        int
            The version of the stream.

        """
        return self._btrdb.ep.streamInfo(self._uuid, True, False)[4]

    def insert(self, data, merge='never'):
        """
        Insert new data in the form (time, value) into the series.

        Inserts a list of new (time, value) tuples into the series. The tuples
        in the list need not be sorted by time. If the arrays are larger than
        appropriate, this function will automatically chunk the inserts. As a
        consequence, the insert is not necessarily atomic, but can be used with
        a very large array.

        Parameters
        ----------
        data: list[tuple[int, float]]
            A list of tuples in which each tuple contains a time (int) and
            value (float) for insertion to the database
        merge: str
            A string describing the merge policy. Valid policies are:
              - 'never': the default, no points are merged
              - 'equal': points are deduplicated if the time and value are equal
              - 'retain': if two points have the same timestamp, the old one is kept
              - 'replace': if two points have the same timestamp, the new one is kept

        Returns
        -------
        int
            The version of the stream after inserting new points.

        """
        i = 0
        version = 0
        while i < len(data):
            thisBatch = data[i:i + INSERT_BATCH_SIZE]
            version = self._btrdb.ep.insert(self._uuid, thisBatch, merge)
            i += INSERT_BATCH_SIZE
        return version

    def _update_tags_collection(self, tags, collection):
        tags = self.tags() if tags is None else tags
        collection = self.collection if collection is None else collection
        if collection is None:
            raise ValueError("collection must be provided to update tags or collection")

        self._btrdb.ep.setStreamTags(
            uu=self.uuid,
            expected=self._property_version,
            tags=tags,
            collection=collection
        )

    def _update_annotations(self, annotations, encoder, replace):
        # make a copy of the annotations to prevent accidental mutable object mutation
        serialized = deepcopy(annotations)
        if encoder is not None:
            serialized = {
                k: json.dumps(v, cls=encoder, indent=None, allow_nan=True)
                for k, v in serialized.items()
            }

        removals = []
        if replace:
            removals = [i for i in self._annotations.keys() if i not in annotations.keys()]

        self._btrdb.ep.setStreamAnnotations(
            uu=self.uuid,
            expected=self._property_version,
            changes=serialized,
            removals=removals
        )

    def update(self, tags=None, annotations=None, collection=None, encoder=AnnotationEncoder, replace=False):
        """
        Updates metadata including tags, annotations, and collection as an
        UPSERT operation.

        By default, the update will only affect the keys and values in the
        specified tags and annotations dictionaries, inserting them if they
        don't exist, or updating the value for the key if it does exist. If
        any of the update arguments (i.e. tags, annotations, collection) are
        None, they will remain unchanged in the database.

        To delete either tags or annotations, you must specify exactly which
        keys and values you want set for the field and set `replace=True`. For
        example:

            >>> annotations, _ = stream.anotations()
            >>> del annotations["key_to_delete"]
            >>> stream.update(annotations=annotations, replace=True)

        This ensures that all of the keys and values for the annotations are
        preserved except for the key to be deleted.

        Parameters
        -----------
        tags : dict, optional
            Specify the tag key/value pairs as a dictionary to upsert or
            replace. If None, the tags  will remain unchanged in the database.
        annotations : dict, optional
            Specify the annotations key/value pairs as a dictionary to upsert
            or replace. If None, the annotations will remain unchanged in the
            database.
        collection : str, optional
            Specify a new collection for the stream. If None, the collection
            will remain unchanged.
        encoder : json.JSONEncoder or None
            JSON encoder class to use for annotation serialization. Set to None
            to prevent JSON encoding of the annotations.
        replace : bool, default: False
            Replace all annotations or tags with the specified dictionaries
            instead of performing the normal upsert operation. Specifying True
            is the only way to remove annotation keys.

        Returns
        -------
        int
            The version of the metadata (separate from the version of the data)
            also known as the "property version".

        """
        if tags is None and annotations is None and collection is None:
            raise ValueError("you must supply a tags, annotations, or collection argument")

        if tags is not None and isinstance(tags, dict) is False:
            raise TypeError("tags must be of type dict")

        if annotations is not None and isinstance(annotations, dict) is False:
            raise TypeError("annotations must be of type dict")

        if collection is not None and isinstance(collection, str) is False:
            raise TypeError("collection must be of type string")

        if tags is not None or collection is not None:
            self._update_tags_collection(tags, collection)
            self.refresh_metadata()

        if annotations is not None:
            self._update_annotations(annotations, encoder, replace)
            self.refresh_metadata()

        return self._property_version

    def delete(self, start, end):
        """
        "Delete" all points between [`start`, `end`)

        "Delete" all points between `start` (inclusive) and `end` (exclusive),
        both in nanoseconds. As BTrDB has persistent multiversioning, the
        deleted points will still exist as part of an older version of the
        stream.

        Parameters
        ----------
        start : int or datetime like object
            The start time in nanoseconds for the range to be deleted. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        end : int or datetime like object
            The end time in nanoseconds for the range to be deleted. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)

        Returns
        -------
        int
            The version of the new stream created

        """
        return self._btrdb.ep.deleteRange(self._uuid, to_nanoseconds(start),
            to_nanoseconds(end))

    def values(self, start, end, version=0):
        """
        Read raw values from BTrDB between time [a, b) in nanoseconds.

        RawValues queries BTrDB for the raw time series data points between
        `start` and `end` time, both in nanoseconds since the Epoch for the
        specified stream `version`.

        Parameters
        ----------
        start : int or datetime like object
            The start time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        end : int or datetime like object
            The end time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        version: int
            The version of the stream to be queried

        Returns
        ------
        list
            Returns a list of tuples containing a RawPoint and the stream
            version (list(tuple(RawPoint,int))).


        Notes
        -----
        Note that the raw data points are the original values at the sensor's
        native sampling rate (assuming the time series represents measurements
        from a sensor). This is the lowest level of data with the finest time
        granularity. In the tree data structure of BTrDB, this data is stored in
        the vector nodes.

        """
        materialized = []
        start = to_nanoseconds(start)
        end = to_nanoseconds(end)

        point_windows = self._btrdb.ep.rawValues(self._uuid, start, end, version)
        for point_list, version in point_windows:
            for point in point_list:
                materialized.append((RawPoint.from_proto(point), version))
        return materialized

    def aligned_windows(self, start, end, pointwidth, version=0):
        """
        Read statistical aggregates of windows of data from BTrDB.

        Query BTrDB for aggregates (or roll ups or windows) of the time series
        with `version` between time `start` (inclusive) and `end` (exclusive) in
        nanoseconds. Each point returned is a statistical aggregate of all the
        raw data within a window of width 2**`pointwidth` nanoseconds. These
        statistical aggregates currently include the mean, minimum, and maximum
        of the data and the count of data points composing the window.

        Note that `start` is inclusive, but `end` is exclusive. That is, results
        will be returned for all windows that start in the interval [start, end).
        If end < start+2^pointwidth you will not get any results. If start and
        end are not powers of two, the bottom pointwidth bits will be cleared.
        Each window will contain statistical summaries of the window.
        Statistical points with count == 0 will be omitted.

        Parameters
        ----------
        start : int or datetime like object
            The start time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        end : int or datetime like object
            The end time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        pointwidth : int
            Specify the number of ns between data points (2**pointwidth)
        version : int
            Version of the stream to query

        Returns
        -------
        tuple
            Returns a tuple containing windows of data.  Each window is a tuple
            containing data tuples.  Each data tuple contains a StatPoint and
            the stream version.

        Notes
        -----
        As the window-width is a power-of-two, it aligns with BTrDB internal
        tree data structure and is faster to execute than `windows()`.
        """
        materialized = []
        start = to_nanoseconds(start)
        end = to_nanoseconds(end)

        windows = self._btrdb.ep.alignedWindows(self._uuid, start, end, pointwidth, version)
        for stat_points, version in windows:
            for point in stat_points:
                materialized.append((StatPoint.from_proto(point), version))

        return tuple(materialized)

    def windows(self, start, end, width, depth=0, version=0):
        """
        Read arbitrarily-sized windows of data from BTrDB.  StatPoint objects
        will be returned representing the data for each window.

        Parameters
        ----------
        start : int or datetime like object
            The start time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        end : int or datetime like object
            The end time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        width : int
            The number of nanoseconds in each window, subject to the depth
            parameter.
        depth : int
            The precision of the window duration as a power of 2 in nanoseconds.
            E.g 30 would make the window duration accurate to roughly 1 second
        version : int
            The version of the stream to query.

        Returns
        -------
        tuple
            Returns a tuple containing windows of data.  Each window is a tuple
            containing data tuples.  Each data tuple contains a StatPoint and
            the stream version (tuple(tuple(StatPoint, int), ...)).

        Notes
        -----
        Windows returns arbitrary precision windows from BTrDB. It is slower
        than AlignedWindows, but still significantly faster than RawValues. Each
        returned window will be `width` nanoseconds long. `start` is inclusive,
        but `end` is exclusive (e.g if end < start+width you will get no
        results). That is, results will be returned for all windows that start
        at a time less than the end timestamp. If (`end` - `start`) is not a
        multiple of width, then end will be decreased to the greatest value less
        than end such that (end - start) is a multiple of `width` (i.e., we set
        end = start + width * floordiv(end - start, width). The `depth`
        parameter is an optimization that can be used to speed up queries on
        fast queries. Each window will be accurate to 2^depth nanoseconds. If
        depth is zero, the results are accurate to the nanosecond. On a dense
        stream for large windows, this accuracy may not be required. For example
        for a window of a day, +- one second may be appropriate, so a depth of
        30 can be specified. This is much faster to execute on the database
        side.

        """
        materialized = []
        start = to_nanoseconds(start)
        end = to_nanoseconds(end)

        windows = self._btrdb.ep.windows(self._uuid, start, end, width, depth, version)
        for stat_points, version in windows:
            for point in stat_points:
                materialized.append((StatPoint.from_proto(point), version))

        return tuple(materialized)

    def nearest(self, time, version, backward=False):
        """
        Finds the closest point in the stream to a specified time.

        Return the point nearest to the specified `time` in nanoseconds since
        Epoch in the stream with `version` while specifying whether to search
        forward or backward in time. If `backward` is false, the returned point
        will be >= `time`. If backward is true, the returned point will be <
        `time`. The version of the stream used to satisfy the query is returned.

        Parameters
        ----------
        time : int or datetime like object
            The time (in nanoseconds since Epoch) to search near (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)
        version : int
            Version of the stream to use in search
        backward : boolean
            True to search backwards from time, else false for forward

        Returns
        -------
        tuple
            The closest data point in the stream and the version of the stream
            the value was retrieved at (tuple(RawPoint, int)).

        """

        try:
            rp, version = self._btrdb.ep.nearest(self._uuid,
                to_nanoseconds(time), version, backward)
        except BTrDBError as exc:
            if exc.code != 401:
                raise
            return None

        return RawPoint.from_proto(rp), version


    def obliterate(self):
        """
        Obliterate deletes a stream from the BTrDB server.  An exception will be
        raised if the stream could not be found.

        Raises
        ------
        BTrDBError [404] stream does not exist
            The stream could not be found in BTrDB

        """
        self._btrdb.ep.obliterate(self._uuid)

    def flush(self):
        """
        Flush writes the stream buffers out to persistent storage.

        """
        self._btrdb.ep.flush(self._uuid)

    def __repr__(self):
        return "<Stream collection={} name={}>".format(self.collection,
            self.name)

##########################################################################
## StreamSet  Classes
##########################################################################

class StreamSetBase(Sequence):
    """
    A lighweight wrapper around a list of stream objects
    """

    def __init__(self, streams):
        self._streams = streams
        self._pinned_versions = None

        self.filters = []
        self.pointwidth = None
        self.width = None
        self.depth = None

    @property
    def allow_window(self):
        return not bool(self.pointwidth or (self.width and self.depth))

    def _latest_versions(self):
        return {s.uuid: s.version() for s in self._streams}


    def pin_versions(self, versions=None):
        """
        Saves the stream versions that future materializations should use.  If
        no pin is requested then the first materialization will automatically
        pin the return versions.  Versions can also be supplied through a dict
        object with key:UUID, value:stream.version().

        Parameters
        ----------
        versions : dict[UUID: int]
            A dict containing the stream UUID and version ints as key/values

        Returns
        -------
        StreamSet
            Returns self

        """
        if versions is not None:
            if not isinstance(versions, dict):
                raise TypeError("`versions` argument must be dict")

            for key in versions.keys():
                if not isinstance(key, uuidlib.UUID):
                    raise TypeError("version keys must be type UUID")


        self._pinned_versions = self._latest_versions() if not versions else versions
        return self

    def versions(self):
        """
        Returns a dict of the stream versions.  These versions are the pinned
        values if previously pinned or the latest stream versions if not
        pinned.


        Parameters
        ----------
        None

        Returns
        -------
        dict
            A dict containing the stream UUID and version ints as key/values

        """
        return self._pinned_versions if self._pinned_versions else self._latest_versions()

    def count(self):
        """
        Compute the total number of points in the streams using filters.

        Computes the total number of points across all streams using the
        specified filters. By default, this returns the latest total count of
        all points in the streams. The count is modified by start and end
        filters or by pinning versions.

        Note that this helper method sums the counts of all StatPoints returned
        by ``aligned_windows``. Because of this the start and end timestamps
        may be adjusted if they are not powers of 2. You can also set the
        pointwidth property for smaller windows of time to ensure that the
        count granularity is captured appropriately.

        Parameters
        ----------
        None

        Returns
        -------
        int
            The total number of points in all streams for the specified filters.
        """
        params = self._params_from_filters()
        start = params.get("start", MINIMUM_TIME)
        end = params.get("end", MAXIMUM_TIME)
        versions = self._pinned_versions if self._pinned_versions else {}
        count = 0

        for s in self._streams:
            version = versions.get(s.uuid, 0)
            count += s.count(start, end, version=version)

        return count

    def earliest(self):
        """
        Returns earliest points of data in streams using available filters.

        Parameters
        ----------
        None

        Returns
        -------
        tuple
            The earliest points of data found among all streams

        """
        earliest = []
        params = self._params_from_filters()
        start = params.get("start", MINIMUM_TIME)

        for s in self._streams:
            version = self.versions()[s.uuid]
            point, _ = s.nearest(start, version=version, backward=False)
            earliest.append(point)

        return tuple(earliest)

    def latest(self):
        """
        Returns latest points of data in the streams using available filters.

        Parameters
        ----------
        None

        Returns
        -------
        tuple
            The latest points of data found among all streams

        """
        latest = []
        params = self._params_from_filters()
        start = params.get("end", MAXIMUM_TIME)

        for s in self._streams:
            version = self.versions()[s.uuid]
            point, _ = s.nearest(start, version=version, backward=True)
            latest.append(point)

        return tuple(latest)

    def current(self):
        """
        Returns the points of data in the streams closest to the current timestamp. If
        the current timestamp is outside of the filtered range of data, a ValueError is
        raised.

        Parameters
        ----------
        None

        Returns
        -------
        tuple
            The latest points of data found among all streams
        """
        latest = []
        params = self._params_from_filters()
        now = currently_as_ns()
        end = params.get("end", None)
        start = params.get("start", None)

        if (end is not None and end <= now) or (start is not None and start > now):
            raise ValueError("current time is not included in filtered stream range")

        for s in self._streams:
            version = self.versions()[s.uuid]
            point, _ = s.nearest(now, version=version, backward=True)
            latest.append(point)

        return tuple(latest)

    def filter(self, start=None, end=None, collection=None, name=None, unit=None,
               tags=None, annotations=None):
        """
        Provides a new StreamSet instance containing stored query parameters and
        stream objects that match filtering criteria.

        The collection, name, and unit arguments will be used to select streams
        from the original StreamSet object.  If a string is supplied, then a
        case-insensitive exact match is used to select streams.  Otherwise, you
        may supply a compiled regex pattern that will be used with `re.search`.

        The tags and annotations arguments expect dictionaries for the desired
        key/value pairs.  Any stream in the original instance that has the exact
        key/values will be included in the new StreamSet instance.

        Parameters
        ----------
        start : int or datetime like object
            the inclusive start of the query (see :func:`btrdb.utils.timez.to_nanoseconds`
            for valid input types)
        end : int or datetime like object
            the exclusive end of the query (see :func:`btrdb.utils.timez.to_nanoseconds`
            for valid input types)
        collection : str or regex
            string for exact (case-insensitive) matching of collection when filtering streams
            or a compiled regex expression for re.search of stream collections.
        name : str or regex
            string for exact (case-insensitive) matching of name when filtering streams
            or a compiled regex expression for re.search of stream names.
        unit : str or regex
            string for exact (case-insensitive) matching of unit when filtering streams
            or a compiled regex expression for re.search of stream units.
        tags : dict
            key/value pairs for filtering streams based on tags
        annotations : dict
            key/value pairs for filtering streams based on annotations

        Returns
        -------
        StreamSet
            a new instance cloned from the original with filters applied

        """

        obj = self.clone()
        if start is not None or end is not None:
            obj.filters.append(StreamFilter(start, end))

        # filter by collection
        if collection is not None:
            if isinstance(collection, RE_PATTERN):
                obj._streams = [s for s in obj._streams for m in [collection.search(s.collection)] if m]
            elif isinstance(collection, str):
                obj._streams = [s for s in obj._streams if s.collection.lower() == collection.lower()]
            else:
                raise TypeError("collection must be string or compiled regex")

        # filter by name
        if name is not None:
            if isinstance(name, RE_PATTERN):
                obj._streams = [s for s in obj._streams for m in [name.search(s.name)] if m]
            elif isinstance(name, str):
                obj._streams = [s for s in obj._streams if s.name.lower() == name.lower()]
            else:
                raise TypeError("name must be string or compiled regex")

        # filter by unit
        if unit is not None:
            if isinstance(unit, RE_PATTERN):
                obj._streams = [s for s in obj._streams for m in [unit.search(s.tags()["unit"])] if m]
            elif isinstance(unit, str):
                obj._streams = [s for s in obj._streams if s.tags().get("unit", "").lower() == unit.lower()]
            else:
                raise TypeError("unit must be string or compiled regex")

        # filter by tags
        if tags:
            # filters if the subset of the tags matches the given tags
            obj._streams = [
                s for s in obj._streams
                if tags.items() <= s.tags().items()
            ]

        # filter by annotations
        if annotations:
            # filters if the subset of the annotations matches the given annotations
            obj._streams = [
                s for s in obj._streams
                if annotations.items() <= s.annotations()[0].items()
            ]

        return obj

    def clone(self):
        """
        Returns a deep copy of the object.  Attributes that cannot be copied
        will be referenced to both objects.

        Parameters
        ----------
        None

        Returns
        -------
        StreamSet
            Returns a new copy of the instance

        """
        protected = ('_streams', )
        clone = self.__class__(self._streams)
        for attr, val in self.__dict__.items():
            if attr not in protected:
                setattr(clone, attr, deepcopy(val))
        return clone

    def windows(self, width, depth):
        """
        Stores the request for a windowing operation when the query is
        eventually materialized.

        Parameters
        ----------
        width : int
            The number of nanoseconds to use for each window size.
        depth : int
            The requested accuracy of the data up to 2^depth nanoseconds.  A
            depth of 0 is accurate to the nanosecond.

        Returns
        -------
        StreamSet
            Returns self


        Notes
        -----
        Windows returns arbitrary precision windows from BTrDB. It is slower
        than aligned_windows, but still significantly faster than values. Each
        returned window will be width nanoseconds long. start is inclusive, but
        end is exclusive (e.g if end < start+width you will get no results).
        That is, results will be returned for all windows that start at a time
        less than the end timestamp. If (end - start) is not a multiple of
        width, then end will be decreased to the greatest value less than end
        such that (end - start) is a multiple of width (i.e., we set end = start
        + width * floordiv(end - start, width). The depth parameter is an
        optimization that can be used to speed up queries on fast queries. Each
        window will be accurate to 2^depth nanoseconds. If depth is zero, the
        results are accurate to the nanosecond. On a dense stream for large
        windows, this accuracy may not be required. For example for a window of
        a day, +- one second may be appropriate, so a depth of 30 can be
        specified. This is much faster to execute on the database side.

        """
        if not self.allow_window:
            raise InvalidOperation("A window operation is already requested")

        # TODO: refactor keeping in mind how exception is raised
        self.width = int(width)
        self.depth = int(depth)
        return self

    def aligned_windows(self, pointwidth):
        """
        Stores the request for an aligned windowing operation when the query is
        eventually materialized.

        Parameters
        ----------
        pointwidth : int
            The length of each returned window as computed by 2^pointwidth.

        Returns
        -------
        StreamSet
            Returns self

        Notes
        -----
        `aligned_windows` reads power-of-two aligned windows from BTrDB. It is
        faster than Windows(). Each returned window will be 2^pointwidth
        nanoseconds long, starting at start. Note that start is inclusive, but
        end is exclusive. That is, results will be returned for all windows that
        start in the interval [start, end). If end < start+2^pointwidth you will
        not get any results. If start and end are not powers of two, the bottom
        pointwidth bits will be cleared. Each window will contain statistical
        summaries of the window. Statistical points with count == 0 will be
        omitted.

        """
        if not self.allow_window:
            raise InvalidOperation("A window operation is already requested")

        self.pointwidth = int(pointwidth)
        return self

    def _streamset_data(self, as_iterators=False):
        """
        Private method to return a list of lists representing the data from each
        stream within the StreamSetself.

        Parameters
        ----------
        as_iterators : bool
            Returns each single stream's data as an iterator.  Defaults to False.
        """
        params = self._params_from_filters()
        versions = self.versions()
        data = []

        if self.pointwidth is not None:
            # create list of stream.aligned_windows data
            params.update({"pointwidth": self.pointwidth})
            for s in self._streams:
                params.update({"version": versions[s.uuid]})
                data.append(s.aligned_windows(**params))


        elif self.width is not None and self.depth is not None:
            # create list of stream.windows data (the windows method should
            # prevent the possibility that only one of these is None)
            params.update({"width": self.width, "depth": self.depth})
            for s in self._streams:
                params.update({"version": versions[s.uuid]})
                data.append(s.windows(**params))

        else:
            # create list of stream.values
            data = [s.values(**params) for s in self._streams]

        if as_iterators:
            return [iter(ii) for ii in data]

        return data


    def rows(self):
        """
        Returns a materialized list of tuples where each tuple contains the
        points from each stream at a unique time.  If a stream has no value for that
        time than None is provided instead of a point object.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of tuples containing a RawPoint (or StatPoint) and the stream
            version (list(tuple(RawPoint, int))).

        """
        result = []
        streamset_data = self._streamset_data(as_iterators=True)
        buffer = PointBuffer(len(self._streams))

        while True:
            streams_empty = True

            # add next values from streams into buffer
            for stream_idx, data in enumerate(streamset_data):

                if buffer.active[stream_idx]:
                    try:
                        point, _ = next(data)
                        buffer.add_point(stream_idx, point)
                        streams_empty = False
                    except StopIteration:
                        buffer.deactivate(stream_idx)
                        continue

            key = buffer.next_key_ready()
            if key:
                result.append(tuple(buffer.pop(key)))

            if streams_empty and len(buffer.keys()) == 0:
                break

        return result

    def _params_from_filters(self):
        params = {}
        for filter in self.filters:
            if filter.start is not None:
                params["start"] = filter.start
            if filter.end is not None:
                params["end"] = filter.end
        return params

    def values_iter(self):
        """
        Must return context object which would then close server cursor on __exit__
        """
        raise NotImplementedError()

    def values(self):
        """
        Returns a fully materialized list of lists for the stream values/points
        """
        result = []
        streamset_data = self._streamset_data()
        for stream_data in streamset_data:
            result.append([point[0] for point in stream_data])

        return result

    def __repr__(self):
        token = "stream" if len(self) == 1 else "streams"
        return "<{}({} {})>".format(
            self.__class__.__name__, len(self._streams), token
        )

    def __str__(self):
        token = "stream" if len(self) == 1 else "streams"
        return "{} with {} {}".format(
            self.__class__.__name__, len(self._streams), token
        )

    def __getitem__(self, item):
        if isinstance(item, str):
            item = uuidlib.UUID(item)

        if isinstance(item, uuidlib.UUID):
            for stream in self._streams:
                if stream.uuid == item:
                    return stream
            raise KeyError("Stream with uuid `{}` not found.".format(str(item)))

        return self._streams[item]

    def __len__(self):
        return len(self._streams)


class StreamSet(StreamSetBase, StreamSetTransformer):
    """
    Public class for a collection of streams
    """
    pass


##########################################################################
## Utility Classes
##########################################################################

class StreamFilter(object):
    """
    Object for storing requested filtering options
    """
    def __init__(self, start=None, end=None):
        self.start = to_nanoseconds(start) if start else None
        self.end = to_nanoseconds(end) if end else None

        if self.start is None and self.end is None:
            raise ValueError("A valid `start` or `end` must be supplied")

        if self.start is not None and self.end is not None and self.start >= self.end:
            raise ValueError("`start` must be strictly less than `end` argument")



# tests.test_conn
# Testing package for the btrdb connection module
#
# Author:   PingThings
# Created:  Wed Jan 02 19:26:20 2019 -0500
#
# For license information, see LICENSE.txt
# ID: test_conn.py [] allen@pingthings.io $

"""
Testing package for the btrdb connection module
"""

##########################################################################
## Imports
##########################################################################

import uuid as uuidlib
import pytest
from unittest.mock import Mock, PropertyMock, patch, call

from btrdb.conn import Connection, BTrDB
from btrdb.endpoint import Endpoint
from btrdb.grpcinterface import btrdb_pb2
from btrdb.exceptions import *

##########################################################################
## Connection Tests
##########################################################################

class TestConnection(object):

    def test_raises_err_invalid_address(self):
        """
        Assert ValueError is raised if address:port is invalidly formatted
        """
        address = "127.0.0.1::4410"
        with pytest.raises(ValueError) as exc:
            conn = Connection(address)
        assert "expecting address:port" in str(exc)


    def test_raises_err_for_apikey_insecure_port(self):
        """
        Assert error is raised if apikey used on insecure port
        """
        address = "127.0.0.1:4410"
        with pytest.raises(ValueError) as exc:
            conn = Connection(address, apikey="abcd")
        assert "cannot use an API key with an insecure" in str(exc)


##########################################################################
## BTrDB Tests
##########################################################################

class TestBTrDB(object):

    ##########################################################################
    ## .streams tests
    ##########################################################################

    def test_streams_raises_err_if_version_not_list(self):
        """
        Assert streams raises TypeError if versions is not list
        """
        db = BTrDB(None)
        with pytest.raises(TypeError) as exc:
            db.streams('0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a', versions="2,2")

        assert "versions argument must be of type list" in str(exc)


    def test_streams_raises_err_if_version_argument_mismatch(self):
        """
        Assert streams raises ValueError if len(identifiers) doesnt match length of versions
        """
        db = BTrDB(None)
        with pytest.raises(ValueError) as exc:
            db.streams('0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a', versions=[2,2])

        assert "versions does not match identifiers" in str(exc)


    def test_streams_stores_versions(self):
        """
        Assert streams correctly stores supplied version info
        """
        db = BTrDB(None)
        uuid1 = uuidlib.UUID('0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a')
        uuid2 = uuidlib.UUID('17dbe387-89ea-42b6-864b-f505cdb483f5')
        versions = [22,44]
        expected = dict(zip([uuid1, uuid2], versions))

        streams = db.streams(uuid1, uuid2, versions=versions)
        assert streams._pinned_versions == expected


    @patch('btrdb.conn.BTrDB.stream_from_uuid')
    def test_streams_recognizes_uuid(self, mock_func):
        """
        Assert streams recognizes uuid strings
        """
        db = BTrDB(None)
        uuid1 = uuidlib.UUID('0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a')
        mock_func.return_value = [1]
        db.streams(uuid1)

        mock_func.assert_called_once()
        assert mock_func.call_args[0][0] == uuid1


    @patch('btrdb.conn.BTrDB.stream_from_uuid')
    def test_streams_recognizes_uuid_string(self, mock_func):
        """
        Assert streams recognizes uuid strings
        """
        db = BTrDB(None)
        uuid1 = '0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a'
        mock_func.return_value = [1]
        db.streams(uuid1)

        mock_func.assert_called_once()
        assert mock_func.call_args[0][0] == uuid1


    @patch('btrdb.conn.BTrDB.streams_in_collection')
    def test_streams_handles_path(self, mock_func):
        """
        Assert streams calls streams_in_collection for collection/name paths
        """
        db = BTrDB(None)
        ident = "zoo/animal/dog"
        mock_func.return_value = [1]
        db.streams(ident, '0d22a53b-e2ef-4e0a-ab89-b2d48fb2592a')

        mock_func.assert_called_once()
        assert mock_func.call_args[0][0] == 'zoo/animal'
        assert mock_func.call_args[1] == {
            'is_collection_prefix': False,
            'tags': {'name': 'dog'}
        }


    @patch('btrdb.conn.BTrDB.streams_in_collection')
    def test_streams_raises_err(self, mock_func):
        """
        Assert streams raises NotFound
        """
        db = BTrDB(None)
        ident = "zoo/animal/dog"

        mock_func.return_value = []
        with pytest.raises(NotFound) as exc:
            db.streams(ident)

        mock_func.return_value = [1,2]
        with pytest.raises(NotFound) as exc:
            db.streams(ident)

        # check that does not raise if one returned
        mock_func.return_value = [1]
        db.streams(ident)


    def test_streams_raises_valueerror(self):
        """
        Assert streams raises ValueError if not uuid, uuid str, or path
        """
        db = BTrDB(None)
        with pytest.raises(ValueError) as exc:
            db.streams(11)


    ##########################################################################
    ## other tests
    ##########################################################################

    def test_info(self):
        """
        Assert info method returns a dict
        """
        serialized = b'\x18\x05*\x055.0.02\x10\n\x0elocalhost:4410'
        info = btrdb_pb2.InfoResponse.FromString(serialized)

        endpoint = Mock(Endpoint)
        endpoint.info = Mock(return_value=info)
        conn = BTrDB(endpoint)

        truth = {
            "majorVersion": 5,
            "build": "5.0.0",
            "proxy": { "proxyEndpoints": ["localhost:4410"], },
        }
        info = conn.info()
        assert info == truth

        # verify RepeatedScalarContainer is converted to list
        assert info["proxy"]["proxyEndpoints"].__class__ == list

    def test_list_collections(self):
        """
        Assert list_collections method works
        """
        endpoint = Mock(Endpoint)
        endpoint.listCollections = Mock(side_effect=[iter([
            ['allen/automated'],
            ['allen/bindings']
        ])])
        conn = BTrDB(endpoint)

        truth = ['allen/automated', 'allen/bindings']
        assert conn.list_collections() == truth

    @patch('btrdb.conn.unpack_stream_descriptor')
    def test_streams_in_collections_args(self, mock_util):
        """
        Assert streams_in_collections correctly sends *collection, tags, annotations
        to the endpoint method
        """
        descriptor = Mock()
        type(descriptor).uuid = PropertyMock(return_value=uuidlib.uuid4().bytes)
        type(descriptor).collection = PropertyMock(return_value="fruit/apple")
        type(descriptor).propertyVersion = PropertyMock(return_value=22)
        mock_util.side_effect = [({"name": "gala"}, {}), ({"name": "fuji"}, {})]

        endpoint = Mock(Endpoint)
        endpoint.lookupStreams = Mock(side_effect=[[[descriptor]], [[descriptor]]])

        conn = BTrDB(endpoint)
        tags = {"unit": "volts"}
        annotations = {"size": "large"}
        streams = conn.streams_in_collection("a", "b", is_collection_prefix=False,
            tags=tags, annotations=annotations)

        assert streams[0].name == "gala"
        assert streams[1].name == "fuji"

        expected = [call('a', False, tags, annotations), call('b', False, tags, annotations)]
        assert endpoint.lookupStreams.call_args_list == expected

    def test_streams_in_collections_no_arg(self):
        """
        Assert streams_in_collections calls lookupStreams if collections not sent
        """
        endpoint = Mock(Endpoint)
        endpoint.lookupStreams.return_value = []

        conn = BTrDB(endpoint)
        annotations = {"size": "large"}
        streams = conn.streams_in_collection(annotations=annotations)
        assert endpoint.lookupStreams.called



# btrdb.transformers
# Value transformation utilities
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: transformers.py [] allen@pingthings.io $

"""
Value transformation utilities
"""

##########################################################################
## Imports
##########################################################################

import csv
import contextlib
from collections import OrderedDict
from warnings import warn

##########################################################################
## Helper Functions
##########################################################################

_STAT_PROPERTIES = ('min', 'mean', 'max', 'count', 'stddev')

def _get_time_from_row(row):
    for item in row:
        if item: return item.time
    raise Exception("Row contains no data")


def _stream_names(streamset, func):
    """
    private convenience function to come up with proper final stream names
    before sending a collection of streams (dataframe, etc.) back to the
    user.
    """
    return tuple(
        func(s) for s in streamset._streams
    )


##########################################################################
## Transform Functions
##########################################################################

def to_series(streamset, datetime64_index=True, agg="mean", name_callable=None):
    """
    Returns a list of Pandas Series objects indexed by time

    Parameters
    ----------
    datetime64_index: bool
        Directs function to convert Series index to np.datetime64[ns] or
        leave as np.int64.

    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to create the Series
        from. Must be one of "min", "mean", "max", "count", or "stddev". This
        argument is ignored if RawPoint values are passed into the function.

    name_callable : lambda, default: lambda s: s.collection + "/" +  s.name
        Sprecify a callable that can be used to determine the series name given a
        Stream object.

    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Please install Pandas to use this transformation function.")

    # TODO: allow this at some future point
    if agg == "all":
        raise AttributeError("cannot use 'all' as aggregate at this time")

    if not callable(name_callable):
        name_callable = lambda s: s.collection + "/" +  s.name


    result = []
    stream_names = _stream_names(streamset, name_callable)

    for idx, output in enumerate(streamset.values()):
        times, values = [], []
        for point in output:
            times.append(point.time)
            if point.__class__.__name__ == "RawPoint":
                values.append(point.value)
            else:
                values.append(getattr(point, agg))

        if datetime64_index:
            times = pd.Index(times, dtype='datetime64[ns]')

        result.append(pd.Series(
            data=values, index=times, name=stream_names[idx]
        ))
    return result


def to_dataframe(streamset, columns=None, agg="mean", name_callable=None):
    """
    Returns a Pandas DataFrame object indexed by time and using the values of a
    stream for each column.

    Parameters
    ----------
    columns: sequence
        column names to use for DataFrame.  Deprecated and not compatible with name_callable.

    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to create the Series
        from. Must be one of "min", "mean", "max", "count", "stddev", or "all". This
        argument is ignored if not using StatPoints.

    name_callable : lambda, default: lambda s: s.collection + "/" +  s.name
        Sprecify a callable that can be used to determine the series name given a
        Stream object.  This is not compatible with agg == "all" at this time


    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Please install Pandas to use this transformation function.")

    # deprecation warning added in v5.8
    if columns:
        warn("the columns argument is deprecated and will be removed in a future release", DeprecationWarning, stacklevel=2)

    # TODO: allow this at some future point
    if agg == "all" and name_callable is not None:
        raise AttributeError("cannot provide name_callable when using 'all' as aggregate at this time")

    # do not allow agg="all" with RawPoints
    if agg == "all" and streamset.allow_window:
        agg=""

    # default arg values
    if not callable(name_callable):
        name_callable = lambda s: s.collection + "/" +  s.name


    df = pd.DataFrame(to_dict(streamset,agg=agg))
    df = df.set_index("time")

    if agg == "all" and not streamset.allow_window:
        stream_names = [[s.collection, s.name, prop] for s in streamset._streams for prop in _STAT_PROPERTIES]
        df.columns=pd.MultiIndex.from_tuples(stream_names)
    else:
        df.columns =  columns if columns else _stream_names(streamset, name_callable)

    return df


def to_array(streamset, agg="mean"):
    """
    Returns a multidimensional numpy array (similar to a list of lists) containing point
    classes.

    Parameters
    ----------
    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to return for the
        arrays. Must be one of "min", "mean", "max", "count", or "stddev". This
        argument is ignored if RawPoint values are passed into the function.

    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Please install Numpy to use this transformation function.")

    # TODO: allow this at some future point
    if agg == "all":
        raise AttributeError("cannot use 'all' as aggregate at this time")

    results = []
    for points in streamset.values():
        segment = []
        for point in points:
            if point.__class__.__name__ == "RawPoint":
                segment.append(point.value)
            else:
                segment.append(getattr(point, agg))
        results.append(segment)
    return np.array(results)


def to_dict(streamset, agg="mean", name_callable=None):
    """
    Returns a list of OrderedDict for each time code with the appropriate
    stream data attached.

    Parameters
    ----------
    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to constrain dict
        keys. Must be one of "min", "mean", "max", "count", or "stddev". This
        argument is ignored if RawPoint values are passed into the function.

    name_callable : lambda, default: lambda s: s.collection + "/" +  s.name
        Sprecify a callable that can be used to determine the series name given a
        Stream object.

    """
    if not callable(name_callable):
        name_callable = lambda s: s.collection + "/" +  s.name

    data = []
    stream_names = _stream_names(streamset, name_callable)

    for row in streamset.rows():
        item = OrderedDict({
            "time": _get_time_from_row(row),
        })
        for idx, col in enumerate(stream_names):
            if row[idx].__class__.__name__ == "RawPoint":
                item[col] = row[idx].value if row[idx] else None
            else:
                if agg == "all":
                    for stat in _STAT_PROPERTIES:
                        item["{}-{}".format(col, stat)] = getattr(row[idx], stat) if row[idx] else None
                else:
                    item[col] = getattr(row[idx], agg) if row[idx] else None
        data.append(item)
    return data


def to_csv(streamset, fobj, dialect=None, fieldnames=None, agg="mean", name_callable=None):
    """
    Saves stream data as a CSV file.

    Parameters
    ----------
    fobj: str or file-like object
        Path to use for saving CSV file or a file-like object to use to write to.

    dialect: csv.Dialect
        CSV dialect object from Python csv module.  See Python's csv module for
        more information.

    fieldnames: sequence
        A sequence of strings to use as fieldnames in the CSV header.  See
        Python's csv module for more information.

    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to return when
        limiting results. Must be one of "min", "mean", "max", "count", or "stddev".
        This argument is ignored if RawPoint values are passed into the function.

    name_callable : lambda, default: lambda s: s.collection + "/" +  s.name
        Sprecify a callable that can be used to determine the series name given a
        Stream object.
    """

    # TODO: allow this at some future point
    if agg == "all":
        raise AttributeError("cannot use 'all' as aggregate at this time")

    if not callable(name_callable):
        name_callable = lambda s: s.collection + "/" +  s.name

    @contextlib.contextmanager
    def open_path_or_file(path_or_file):
        if isinstance(path_or_file, str):
            f = file_to_close = open(path_or_file, 'w')
        else:
            f = path_or_file
            file_to_close = None
        try:
            yield f
        finally:
            if file_to_close:
                file_to_close.close()

    with open_path_or_file(fobj) as csvfile:
        stream_names = _stream_names(streamset, name_callable)
        fieldnames = fieldnames if fieldnames else ["time"] + list(stream_names)

        writer = csv.DictWriter(csvfile, fieldnames=fieldnames, dialect=dialect)
        writer.writeheader()

        for item in to_dict(streamset, agg=agg):
            writer.writerow(item)


def to_table(streamset, agg="mean", name_callable=None):
    """
    Returns string representation of the data in tabular form using the tabulate
    library.

    Parameters
    ----------
    agg : str, default: "mean"
        Specify the StatPoint field (e.g. aggregating function) to create the Series
        from. Must be one of "min", "mean", "max", "count", or "stddev". This
        argument is ignored if RawPoint values are passed into the function.

    name_callable : lambda, default: lambda s: s.collection + "/" +  s.name
        Sprecify a callable that can be used to determine the column name given a
        Stream object.

    """
    try:
        from tabulate import tabulate
    except ImportError:
        raise ImportError("Please install tabulate to use this transformation function.")

    # TODO: allow this at some future point
    if agg == "all":
        raise AttributeError("cannot use 'all' as aggregate at this time")

    if not callable(name_callable):
        name_callable = lambda s: s.collection + "/" +  s.name

    return tabulate(streamset.to_dict(agg=agg, name_callable=name_callable), headers="keys")


##########################################################################
## Transform Classes
##########################################################################

class StreamSetTransformer(object):
    """
    Base class for StreamSet or Stream transformations
    """
    to_dict = to_dict
    to_array = to_array
    to_series = to_series
    to_dataframe = to_dataframe

    to_csv = to_csv
    to_table = to_table



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import btrdb.grpcinterface.btrdb_pb2 as btrdb__pb2


class BTrDBStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RawValues = channel.unary_stream(
        '/v5api.BTrDB/RawValues',
        request_serializer=btrdb__pb2.RawValuesParams.SerializeToString,
        response_deserializer=btrdb__pb2.RawValuesResponse.FromString,
        )
    self.AlignedWindows = channel.unary_stream(
        '/v5api.BTrDB/AlignedWindows',
        request_serializer=btrdb__pb2.AlignedWindowsParams.SerializeToString,
        response_deserializer=btrdb__pb2.AlignedWindowsResponse.FromString,
        )
    self.Windows = channel.unary_stream(
        '/v5api.BTrDB/Windows',
        request_serializer=btrdb__pb2.WindowsParams.SerializeToString,
        response_deserializer=btrdb__pb2.WindowsResponse.FromString,
        )
    self.StreamInfo = channel.unary_unary(
        '/v5api.BTrDB/StreamInfo',
        request_serializer=btrdb__pb2.StreamInfoParams.SerializeToString,
        response_deserializer=btrdb__pb2.StreamInfoResponse.FromString,
        )
    self.SetStreamAnnotations = channel.unary_unary(
        '/v5api.BTrDB/SetStreamAnnotations',
        request_serializer=btrdb__pb2.SetStreamAnnotationsParams.SerializeToString,
        response_deserializer=btrdb__pb2.SetStreamAnnotationsResponse.FromString,
        )
    self.SetStreamTags = channel.unary_unary(
        '/v5api.BTrDB/SetStreamTags',
        request_serializer=btrdb__pb2.SetStreamTagsParams.SerializeToString,
        response_deserializer=btrdb__pb2.SetStreamTagsResponse.FromString,
        )
    self.Create = channel.unary_unary(
        '/v5api.BTrDB/Create',
        request_serializer=btrdb__pb2.CreateParams.SerializeToString,
        response_deserializer=btrdb__pb2.CreateResponse.FromString,
        )
    self.ListCollections = channel.unary_stream(
        '/v5api.BTrDB/ListCollections',
        request_serializer=btrdb__pb2.ListCollectionsParams.SerializeToString,
        response_deserializer=btrdb__pb2.ListCollectionsResponse.FromString,
        )
    self.LookupStreams = channel.unary_stream(
        '/v5api.BTrDB/LookupStreams',
        request_serializer=btrdb__pb2.LookupStreamsParams.SerializeToString,
        response_deserializer=btrdb__pb2.LookupStreamsResponse.FromString,
        )
    self.Nearest = channel.unary_unary(
        '/v5api.BTrDB/Nearest',
        request_serializer=btrdb__pb2.NearestParams.SerializeToString,
        response_deserializer=btrdb__pb2.NearestResponse.FromString,
        )
    self.Changes = channel.unary_stream(
        '/v5api.BTrDB/Changes',
        request_serializer=btrdb__pb2.ChangesParams.SerializeToString,
        response_deserializer=btrdb__pb2.ChangesResponse.FromString,
        )
    self.Insert = channel.unary_unary(
        '/v5api.BTrDB/Insert',
        request_serializer=btrdb__pb2.InsertParams.SerializeToString,
        response_deserializer=btrdb__pb2.InsertResponse.FromString,
        )
    self.Delete = channel.unary_unary(
        '/v5api.BTrDB/Delete',
        request_serializer=btrdb__pb2.DeleteParams.SerializeToString,
        response_deserializer=btrdb__pb2.DeleteResponse.FromString,
        )
    self.Info = channel.unary_unary(
        '/v5api.BTrDB/Info',
        request_serializer=btrdb__pb2.InfoParams.SerializeToString,
        response_deserializer=btrdb__pb2.InfoResponse.FromString,
        )
    self.FaultInject = channel.unary_unary(
        '/v5api.BTrDB/FaultInject',
        request_serializer=btrdb__pb2.FaultInjectParams.SerializeToString,
        response_deserializer=btrdb__pb2.FaultInjectResponse.FromString,
        )
    self.Flush = channel.unary_unary(
        '/v5api.BTrDB/Flush',
        request_serializer=btrdb__pb2.FlushParams.SerializeToString,
        response_deserializer=btrdb__pb2.FlushResponse.FromString,
        )
    self.Obliterate = channel.unary_unary(
        '/v5api.BTrDB/Obliterate',
        request_serializer=btrdb__pb2.ObliterateParams.SerializeToString,
        response_deserializer=btrdb__pb2.ObliterateResponse.FromString,
        )
    self.GetMetadataUsage = channel.unary_unary(
        '/v5api.BTrDB/GetMetadataUsage',
        request_serializer=btrdb__pb2.MetadataUsageParams.SerializeToString,
        response_deserializer=btrdb__pb2.MetadataUsageResponse.FromString,
        )
    self.GenerateCSV = channel.unary_stream(
        '/v5api.BTrDB/GenerateCSV',
        request_serializer=btrdb__pb2.GenerateCSVParams.SerializeToString,
        response_deserializer=btrdb__pb2.GenerateCSVResponse.FromString,
        )
    self.SQLQuery = channel.unary_stream(
        '/v5api.BTrDB/SQLQuery',
        request_serializer=btrdb__pb2.SQLQueryParams.SerializeToString,
        response_deserializer=btrdb__pb2.SQLQueryResponse.FromString,
        )


class BTrDBServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RawValues(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AlignedWindows(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Windows(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StreamInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetStreamAnnotations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetStreamTags(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Create(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListCollections(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LookupStreams(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Nearest(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Changes(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Insert(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Delete(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Info(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FaultInject(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Flush(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Obliterate(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMetadataUsage(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GenerateCSV(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SQLQuery(self, request, context):
    """rpc SetCompactionConfig(SetCompactionConfigParams) returns (SetCompactionConfigResponse);
    rpc GetCompactionConfig(GetCompactionConfigParams) returns (GetCompactionConfigResponse);
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_BTrDBServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RawValues': grpc.unary_stream_rpc_method_handler(
          servicer.RawValues,
          request_deserializer=btrdb__pb2.RawValuesParams.FromString,
          response_serializer=btrdb__pb2.RawValuesResponse.SerializeToString,
      ),
      'AlignedWindows': grpc.unary_stream_rpc_method_handler(
          servicer.AlignedWindows,
          request_deserializer=btrdb__pb2.AlignedWindowsParams.FromString,
          response_serializer=btrdb__pb2.AlignedWindowsResponse.SerializeToString,
      ),
      'Windows': grpc.unary_stream_rpc_method_handler(
          servicer.Windows,
          request_deserializer=btrdb__pb2.WindowsParams.FromString,
          response_serializer=btrdb__pb2.WindowsResponse.SerializeToString,
      ),
      'StreamInfo': grpc.unary_unary_rpc_method_handler(
          servicer.StreamInfo,
          request_deserializer=btrdb__pb2.StreamInfoParams.FromString,
          response_serializer=btrdb__pb2.StreamInfoResponse.SerializeToString,
      ),
      'SetStreamAnnotations': grpc.unary_unary_rpc_method_handler(
          servicer.SetStreamAnnotations,
          request_deserializer=btrdb__pb2.SetStreamAnnotationsParams.FromString,
          response_serializer=btrdb__pb2.SetStreamAnnotationsResponse.SerializeToString,
      ),
      'SetStreamTags': grpc.unary_unary_rpc_method_handler(
          servicer.SetStreamTags,
          request_deserializer=btrdb__pb2.SetStreamTagsParams.FromString,
          response_serializer=btrdb__pb2.SetStreamTagsResponse.SerializeToString,
      ),
      'Create': grpc.unary_unary_rpc_method_handler(
          servicer.Create,
          request_deserializer=btrdb__pb2.CreateParams.FromString,
          response_serializer=btrdb__pb2.CreateResponse.SerializeToString,
      ),
      'ListCollections': grpc.unary_stream_rpc_method_handler(
          servicer.ListCollections,
          request_deserializer=btrdb__pb2.ListCollectionsParams.FromString,
          response_serializer=btrdb__pb2.ListCollectionsResponse.SerializeToString,
      ),
      'LookupStreams': grpc.unary_stream_rpc_method_handler(
          servicer.LookupStreams,
          request_deserializer=btrdb__pb2.LookupStreamsParams.FromString,
          response_serializer=btrdb__pb2.LookupStreamsResponse.SerializeToString,
      ),
      'Nearest': grpc.unary_unary_rpc_method_handler(
          servicer.Nearest,
          request_deserializer=btrdb__pb2.NearestParams.FromString,
          response_serializer=btrdb__pb2.NearestResponse.SerializeToString,
      ),
      'Changes': grpc.unary_stream_rpc_method_handler(
          servicer.Changes,
          request_deserializer=btrdb__pb2.ChangesParams.FromString,
          response_serializer=btrdb__pb2.ChangesResponse.SerializeToString,
      ),
      'Insert': grpc.unary_unary_rpc_method_handler(
          servicer.Insert,
          request_deserializer=btrdb__pb2.InsertParams.FromString,
          response_serializer=btrdb__pb2.InsertResponse.SerializeToString,
      ),
      'Delete': grpc.unary_unary_rpc_method_handler(
          servicer.Delete,
          request_deserializer=btrdb__pb2.DeleteParams.FromString,
          response_serializer=btrdb__pb2.DeleteResponse.SerializeToString,
      ),
      'Info': grpc.unary_unary_rpc_method_handler(
          servicer.Info,
          request_deserializer=btrdb__pb2.InfoParams.FromString,
          response_serializer=btrdb__pb2.InfoResponse.SerializeToString,
      ),
      'FaultInject': grpc.unary_unary_rpc_method_handler(
          servicer.FaultInject,
          request_deserializer=btrdb__pb2.FaultInjectParams.FromString,
          response_serializer=btrdb__pb2.FaultInjectResponse.SerializeToString,
      ),
      'Flush': grpc.unary_unary_rpc_method_handler(
          servicer.Flush,
          request_deserializer=btrdb__pb2.FlushParams.FromString,
          response_serializer=btrdb__pb2.FlushResponse.SerializeToString,
      ),
      'Obliterate': grpc.unary_unary_rpc_method_handler(
          servicer.Obliterate,
          request_deserializer=btrdb__pb2.ObliterateParams.FromString,
          response_serializer=btrdb__pb2.ObliterateResponse.SerializeToString,
      ),
      'GetMetadataUsage': grpc.unary_unary_rpc_method_handler(
          servicer.GetMetadataUsage,
          request_deserializer=btrdb__pb2.MetadataUsageParams.FromString,
          response_serializer=btrdb__pb2.MetadataUsageResponse.SerializeToString,
      ),
      'GenerateCSV': grpc.unary_stream_rpc_method_handler(
          servicer.GenerateCSV,
          request_deserializer=btrdb__pb2.GenerateCSVParams.FromString,
          response_serializer=btrdb__pb2.GenerateCSVResponse.SerializeToString,
      ),
      'SQLQuery': grpc.unary_stream_rpc_method_handler(
          servicer.SQLQuery,
          request_deserializer=btrdb__pb2.SQLQueryParams.FromString,
          response_serializer=btrdb__pb2.SQLQueryResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'v5api.BTrDB', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2017 Sam Kumar <samkumar@berkeley.edu>
# Copyright (c) 2017 Michael P Andersen <m.andersen@cs.berkeley.edu>
# Copyright (c) 2017 University of California, Berkeley
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of California, Berkeley nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNERS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from btrdb.grpcinterface import btrdb_pb2
from btrdb.grpcinterface import btrdb_pb2_grpc
from btrdb.point import RawPoint
from btrdb.exceptions import BTrDBError
from btrdb.utils.general import unpack_stream_descriptor

class Endpoint(object):
    def __init__(self, channel):
        self.stub = btrdb_pb2_grpc.BTrDBStub(channel)

    def rawValues(self, uu, start, end, version = 0):
        params = btrdb_pb2.RawValuesParams(uuid = uu.bytes, start = start, end = end, versionMajor = version)
        for result in self.stub.RawValues(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.values, result.versionMajor

    def alignedWindows(self, uu, start, end, pointwidth, version = 0):
        params = btrdb_pb2.AlignedWindowsParams(uuid = uu.bytes, start = start, end = end, versionMajor = version, pointWidth = int(pointwidth))
        for result in self.stub.AlignedWindows(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.values, result.versionMajor

    def windows(self, uu, start, end, width, depth, version = 0):
        params = btrdb_pb2.WindowsParams(uuid = uu.bytes, start = start, end = end, versionMajor = version, width = width, depth = depth)
        for result in self.stub.Windows(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.values, result.versionMajor

    def streamInfo(self, uu, omitDescriptor, omitVersion):
        params = btrdb_pb2.StreamInfoParams(uuid = uu.bytes, omitVersion = omitVersion, omitDescriptor = omitDescriptor)
        result = self.stub.StreamInfo(params)
        desc = result.descriptor
        BTrDBError.checkProtoStat(result.stat)
        tagsanns = unpack_stream_descriptor(desc)
        return desc.collection, desc.propertyVersion, tagsanns[0], tagsanns[1], result.versionMajor

    def obliterate(self, uu):
        params = btrdb_pb2.ObliterateParams(uuid = uu.bytes)
        result = self.stub.Obliterate(params)
        BTrDBError.checkProtoStat(result.stat)

    def setStreamAnnotations(self, uu, expected, changes, removals):
        annkvlist = []
        for k, v in changes.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value = v)
            kv = btrdb_pb2.KeyOptValue(key = k, val = ov)
            annkvlist.append(kv)
        params = btrdb_pb2.SetStreamAnnotationsParams(uuid=uu.bytes, expectedPropertyVersion=expected, changes=annkvlist, removals=removals)
        result = self.stub.SetStreamAnnotations(params)
        BTrDBError.checkProtoStat(result.stat)

    def setStreamTags(self, uu, expected, tags, collection):
        tag_data = []
        for k, v in tags.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value = v)
            kv = btrdb_pb2.KeyOptValue(key = k, val = ov)
            tag_data.append(kv)
        params = btrdb_pb2.SetStreamTagsParams(uuid=uu.bytes, expectedPropertyVersion=expected, tags=tag_data, collection=collection)
        result = self.stub.SetStreamTags(params)
        BTrDBError.checkProtoStat(result.stat)

    def create(self, uu, collection, tags, annotations):
        tagkvlist = []
        for k, v in tags.items():
            kv = btrdb_pb2.KeyOptValue(key = k, val = btrdb_pb2.OptValue(value=v))
            tagkvlist.append(kv)
        annkvlist = []
        for k, v in annotations.items():
            kv = btrdb_pb2.KeyOptValue(key = k, val = btrdb_pb2.OptValue(value=v))
            annkvlist.append(kv)
        params = btrdb_pb2.CreateParams(uuid = uu.bytes, collection = collection, tags = tagkvlist, annotations = annkvlist)
        result = self.stub.Create(params)
        BTrDBError.checkProtoStat(result.stat)

    def listCollections(self, prefix):
        """
        Returns a generator for windows of collection paths matching search

        Yields
        ------
        collection paths : list[str]
        """
        params = btrdb_pb2.ListCollectionsParams(prefix=prefix)
        for msg in self.stub.ListCollections(params):
            BTrDBError.checkProtoStat(msg.stat)
            yield msg.collections

    def lookupStreams(self, collection, isCollectionPrefix, tags, annotations):
        tagkvlist = []
        for k, v in tags.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value = v)
            kv = btrdb_pb2.KeyOptValue(key = k, val = ov)
            tagkvlist.append(kv)
        annkvlist = []
        for k, v in annotations.items():
            if v is None:
                ov = None
            else:
                if isinstance(v, str):
                    v = v.encode("utf-8")
                ov = btrdb_pb2.OptValue(value = v)
            kv = btrdb_pb2.KeyOptValue(key = k, val = ov)
            annkvlist.append(kv)
        params = btrdb_pb2.LookupStreamsParams(collection = collection, isCollectionPrefix = isCollectionPrefix, tags = tagkvlist, annotations = annkvlist)
        for result in self.stub.LookupStreams(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.results

    def nearest(self, uu, time, version, backward):
        params = btrdb_pb2.NearestParams(uuid = uu.bytes, time = time, versionMajor = version, backward = backward)
        result = self.stub.Nearest(params)
        BTrDBError.checkProtoStat(result.stat)
        return result.value, result.versionMajor

    def changes(self, uu, fromVersion, toVersion, resolution):
        params = btrdb_pb2.ChangesParams(uuid = uu.bytes, fromMajor = fromVersion, toMajor = toVersion, resolution = resolution)
        for result in self.stub.Changes(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.ranges, result.versionMajor

    def insert(self, uu, values, policy):
        policy_map = {
            'never': btrdb_pb2.MergePolicy.NEVER,
            'equal': btrdb_pb2.MergePolicy.EQUAL,
            'retain': btrdb_pb2.MergePolicy.RETAIN,
            'replace': btrdb_pb2.MergePolicy.REPLACE,
        }
        protoValues = RawPoint.to_proto_list(values)
        params = btrdb_pb2.InsertParams(uuid = uu.bytes, sync = False, values = protoValues, merge_policy = policy_map[policy])
        result = self.stub.Insert(params)
        BTrDBError.checkProtoStat(result.stat)
        return result.versionMajor

    def deleteRange(self, uu, start, end):
        params = btrdb_pb2.DeleteParams(uuid = uu.bytes, start = start, end = end)
        result = self.stub.Delete(params)
        BTrDBError.checkProtoStat(result.stat)
        return result.versionMajor

    def info(self):
        params = btrdb_pb2.InfoParams()
        result = self.stub.Info(params)
        BTrDBError.checkProtoStat(result.stat)
        return result

    def faultInject(self, typ, args):
        params = btrdb_pb2.FaultInjectParams(type = typ, params = args)
        result = self.stub.FaultInject(params)
        BTrDBError.checkProtoStat(result.stat)
        return result.rv

    def flush(self, uu):
        params = btrdb_pb2.FlushParams(uuid = uu.bytes)
        result = self.stub.Flush(params)
        BTrDBError.checkProtoStat(result.stat)

    def getMetadataUsage(self, prefix):
        params = btrdb_pb2.MetadataUsageParams(prefix = prefix)
        result = self.stub.GetMetadataUsage(params)
        BTrDBError.checkProtoStat(result.stat)
        return result.tags, result.annotations

    def generateCSV(self, queryType, start, end, width, depth, includeVersions, *streams):
        protoStreams = [btrdb_pb2.StreamCSVConfig(version = stream[0],
                                                  label = stream[1],
                                                  uuid = stream[2].bytes)
                        for stream in streams]
        params = btrdb_pb2.GenerateCSVParams(queryType = queryType.to_proto(),
                                             startTime = start,
                                             endTime = end,
                                             windowSize = width,
                                             depth = depth,
                                             includeVersions = includeVersions,
                                             streams = protoStreams)
        result = self.stub.GenerateCSV(params)
        for result in self.stub.GenerateCSV(params):
            BTrDBError.checkProtoStat(result.stat)
            yield result.row

    def sql_query(self, stmt, params=[]):
        request = btrdb_pb2.SQLQueryParams(query=stmt, params=params)
        for page in self.stub.SQLQuery(request):
            BTrDBError.checkProtoStat(page.stat)
            yield page.SQLQueryRow



# tests.test_base
# Testing package for the btrdb library.
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: test_base.py [] allen@pingthings.io $

"""
Testing package for the btrdb database library.
"""

##########################################################################
## Imports
##########################################################################

import os
import pytest
from unittest.mock import patch

from btrdb import connect, __version__, BTRDB_ENDPOINTS, BTRDB_API_KEY
from btrdb.exceptions import ConnectionError


##########################################################################
## Initialization Tests
##########################################################################

class TestConnect(object):

    def setup(self):
        for env in ["BTRDB_ENDPOINTS", "BTRDB_PROFILE", "BTRDB_API_KEY"]:
            try:
                del os.environ[env]
            except KeyError:
                pass

    @patch('btrdb.credentials')
    def test_raises_err_with_no_inputs(self, mock_credentials):
        """
        Assert ConnectionError raised if no connection arg, ENV, profile
        """
        mock_credentials.return_value = {}
        msg = "Could not determine credentials to use."
        with pytest.raises(ConnectionError, match=msg):
            connect()

    def test_raises_err_if_both_profile_and_credentials(self):
        """
        Assert error is raised if both profile and credentials are sent
        """
        with pytest.raises(ValueError):
            connect("192.168.1.100:4410",None,"default")

    @patch('btrdb.utils.credentials.credentials_by_profile')
    @patch('btrdb.utils.credentials.credentials_by_env')
    @patch('btrdb._connect')
    def test_uses_args_over_env(self, mock_connect, mock_credentials_by_env, mock_credentials_by_profile):
        """
        Assert uses connect args over env
        """
        mock_credentials_by_profile.return_value = {}
        mock_credentials_by_env.return_value = {
            "endpoints": "a", "apikey": "b"
        }
        connect("cat", "dog")
        mock_connect.assert_called_once_with(endpoints="cat", apikey="dog")

    @patch('btrdb.utils.credentials.credentials_by_profile')
    @patch('btrdb.utils.credentials.credentials_by_env')
    @patch('btrdb._connect')
    def test_uses_env_over_profile(self, mock_connect, mock_credentials_by_env, mock_credentials_by_profile):
        """
        Assert connect uses env over profile info
        """
        mock_credentials_by_profile.return_value = {
            "endpoints": "a", "apikey": "b"
        }
        mock_credentials_by_env.return_value = {
            "endpoints": "c", "apikey": "d"
        }
        connect()
        mock_connect.assert_called_once_with(endpoints="c", apikey="d")

    @patch('btrdb.utils.credentials.credentials_by_profile')
    @patch('btrdb.Connection')
    def test_connect_with_env(self, mock_conn, mock_credentials_by_profile):
        """
        Assert connect uses ENV variables
        """
        mock_credentials_by_profile.return_value = {}
        address = "127.0.0.1:4410"
        os.environ[BTRDB_ENDPOINTS] = address

        btrdb = connect()
        mock_conn.assert_called_once_with(address, apikey=None)
        mock_conn.reset_mock()

        apikey = "abcd"
        os.environ[BTRDB_API_KEY] = apikey
        btrdb = connect()
        mock_conn.assert_called_once_with(address, apikey=apikey)



# btrdb
# Package for the btrdb database library.
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: __init__.py [] allen@pingthings.io $

"""
Package for the btrdb database library.
"""

##########################################################################
## Imports
##########################################################################

from btrdb.conn import Connection, BTrDB
from btrdb.endpoint import Endpoint
from btrdb.exceptions import ConnectionError
from btrdb.version import get_version
from btrdb.utils.credentials import credentials_by_profile, credentials
from btrdb.stream import MINIMUM_TIME, MAXIMUM_TIME

##########################################################################
## Module Variables
##########################################################################

__version__ = get_version()

BTRDB_ENDPOINTS = "BTRDB_ENDPOINTS"
BTRDB_API_KEY = "BTRDB_API_KEY"
BTRDB_PROFILE = "BTRDB_PROFILE"


##########################################################################
## Functions
##########################################################################

def _connect(endpoints=None, apikey=None):
    return BTrDB(Endpoint(Connection(endpoints, apikey=apikey).channel))

def connect(conn_str=None, apikey=None, profile=None):
    """
    Connect to a BTrDB server.

    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. `192.168.1.1:4411`.
        If set to None, will look in the environment variable `$BTRDB_ENDPOINTS`
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable `$BTRDB_API_KEY`.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        `~/.predictivegrid/credentials.yaml`.

    Returns
    -------
    db : BTrDB
        An instance of the BTrDB context to directly interact with the database.

    """
    # do not allow user to provide both address and profile
    if conn_str and profile:
        raise ValueError("Received both conn_str and profile arguments.")

    # use specific profile if requested
    if profile:
        return _connect(**credentials_by_profile(profile))

    # resolve credentials using combination of arguments, env
    creds = credentials(conn_str, apikey)
    if "endpoints" in creds:
        return _connect(**creds)

    raise ConnectionError("Could not determine credentials to use.")




from functools import partial

import ray

import btrdb
from btrdb.conn import BTrDB

def register_serializer(conn_str=None, apikey=None, profile=None):
    """
    Register serializer for BTrDB Object
    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. `192.168.1.1:4411`.
        If set to None, will look in the environment variable `$BTRDB_ENDPOINTS`
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable `$BTRDB_API_KEY`.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        `~/.predictivegrid/credentials.yaml`.
    """
    ray.register_custom_serializer(
    BTrDB, serializer=btrdb_serializer, deserializer=partial(btrdb_deserializer, conn_str=conn_str, apikey=apikey, profile=profile))

def btrdb_serializer(_):
    """
    sererialize function
    """
    return None

def btrdb_deserializer(_, conn_str=None, apikey=None, profile=None):
    """
    deserialize function
    
    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. `192.168.1.1:4411`.
        If set to None, will look in the environment variable `$BTRDB_ENDPOINTS`
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable `$BTRDB_API_KEY`.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        `~/.predictivegrid/credentials.yaml`.
    Returns
    -------
    db : BTrDB
        An instance of the BTrDB context to directly interact with the database.
    """
    return btrdb.connect(conn_str=conn_str, apikey=apikey, profile=profile)



# btrdb.conn
# Connection related objects for the BTrDB library
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: conn.py [] allen@pingthings.io $

"""
Connection related objects for the BTrDB library
"""

##########################################################################
## Imports
##########################################################################

import os
import re
import json
import uuid as uuidlib

import grpc
from grpc._cython.cygrpc import CompressionAlgorithm

from btrdb.stream import Stream, StreamSet
from btrdb.utils.general import unpack_stream_descriptor
from btrdb.utils.conversion import to_uuid
from btrdb.exceptions import NotFound, InvalidOperation

##########################################################################
## Module Variables
##########################################################################

MIN_TIME = -(16 << 56)
MAX_TIME = 48 << 56
MAX_POINTWIDTH = 63


##########################################################################
## Classes
##########################################################################

class Connection(object):

    def __init__(self, addrportstr, apikey=None):
        """
        Connects to a BTrDB server

        Parameters
        ----------
        addrportstr: str
            The address of the cluster to connect to, e.g 123.123.123:4411
        apikey: str
            The option API key to authenticate requests

        """
        addrport = addrportstr.split(":", 2)
        chan_ops = [('grpc.default_compression_algorithm', CompressionAlgorithm.gzip)]

        if len(addrport) != 2:
            raise ValueError("expecting address:port")

        if addrport[1] == "4411":
            # grpc bundles its own CA certs which will work for all normal SSL
            # certificates but will fail for custom CA certs. Allow the user
            # to specify a CA bundle via env var to overcome this
            ca_bundle = os.getenv("BTRDB_CA_BUNDLE","")
            if ca_bundle != "":
                with open(ca_bundle, "rb") as f:
                    contents = f.read()
            else:
                contents = None

            if apikey is None:
                self.channel = grpc.secure_channel(
                    addrportstr,
                    grpc.ssl_channel_credentials(contents),
                    options=chan_ops
                )
            else:
                self.channel = grpc.secure_channel(
                    addrportstr,
                    grpc.composite_channel_credentials(
                        grpc.ssl_channel_credentials(contents),
                        grpc.access_token_call_credentials(apikey)
                    ),
                    options=chan_ops
                )
        else:
            if apikey is not None:
                raise ValueError("cannot use an API key with an insecure (port 4410) BTrDB API. Try port 4411")
            self.channel = grpc.insecure_channel(addrportstr, chan_ops)




class BTrDB(object):
    """
    The primary server connection object for communicating with a BTrDB server.
    """

    def __init__(self, endpoint):
        self.ep = endpoint

    def query(self, stmt, params=[]):
        """
        Performs a SQL query on the database metadata and returns a list of
        dictionaries from the resulting cursor.

        Parameters
        ----------
        stmt: str
            a SQL statement to be executed on the BTrDB metadata.  Available
            tables are noted below.  To sanitize inputs use a `$1` style parameter such as
            `select * from streams where name = $1 or name = $2`.
        params: list or tuple
            a list of parameter values to be sanitized and interpolated into the
            SQL statement. Using parameters forces value/type checking and is
            considered a best practice at the very least.

        Returns
        -------
        list
            a list of dictionary object representing the cursor results.


        Notes
        -------
        Parameters will be inserted into the SQL statement as noted by the
        paramter number such as `$1`, `$2`, or `$3`.  The `streams` table is
        available for `SELECT` statements only.

        See https://btrdb.readthedocs.io/en/latest/ for more info.
        """
        return [
            json.loads(row.decode("utf-8"))
            for page in self.ep.sql_query(stmt, params)
            for row in page
        ]


    def streams(self, *identifiers, versions=None, is_collection_prefix=False):
        """
        Returns a StreamSet object with BTrDB streams from the supplied
        identifiers.  If any streams cannot be found matching the identifier
        than StreamNotFoundError will be returned.

        Parameters
        ----------
        identifiers: str or UUID
            a single item or iterable of items which can be used to query for
            streams.  identiers are expected to be UUID as string, UUID as UUID,
            or collection/name string.

        versions: list[int]
            a single or iterable of version numbers to match the identifiers

        """
        if versions is not None and not isinstance(versions, list):
            raise TypeError("versions argument must be of type list")

        if versions and len(versions) != len(identifiers):
            raise ValueError("number of versions does not match identifiers")

        streams = []
        for ident in identifiers:
            if isinstance(ident, uuidlib.UUID):
                streams.append(self.stream_from_uuid(ident))
                continue

            if isinstance(ident, str):
                # attempt UUID lookup
                pattern = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
                if re.match(pattern, ident):
                    streams.append(self.stream_from_uuid(ident))
                    continue

                # attempt collection/name lookup
                if "/" in ident:
                    parts = ident.split("/")
                    found = self.streams_in_collection(
                        "/".join(parts[:-1]),
                        is_collection_prefix=is_collection_prefix,
                        tags={"name": parts[-1]}
                    )
                    if len(found) == 1:
                        streams.append(found[0])
                        continue
                    raise NotFound(f"Could not identify stream `{ident}`")

            raise ValueError(f"Could not identify stream based on `{ident}`.  Identifier must be UUID or collection/name.")


        obj = StreamSet(streams)

        if versions:
            version_dict = {streams[idx].uuid: versions[idx] for idx in range(len(versions))}
            obj.pin_versions(version_dict)

        return obj

    def stream_from_uuid(self, uuid):
        """
        Creates a stream handle to the BTrDB stream with the UUID `uuid`. This
        method does not check whether a stream with the specified UUID exists.
        It is always good form to check whether the stream existed using
        `stream.exists()`.


        Parameters
        ----------
        uuid: UUID
            The uuid of the requested stream.

        Returns
        -------
        Stream
            instance of Stream class or None

        """
        return Stream(self, to_uuid(uuid))

    def create(self, uuid, collection, tags=None, annotations=None):
        """
        Tells BTrDB to create a new stream with UUID `uuid` in `collection` with specified `tags` and `annotations`.

        Parameters
        ----------
        uuid: UUID
            The uuid of the requested stream.

        Returns
        -------
        Stream
            instance of Stream class
        """

        if tags is None:
            tags = {}

        if annotations is None:
            annotations = {}

        self.ep.create(uuid, collection, tags, annotations)
        return Stream(self, uuid,
            known_to_exist=True,
            collection=collection,
            tags=tags.copy(),
            annotations=annotations.copy(),
            property_version=0
        )

    def info(self):
        """
        Returns information about the connected BTrDB srerver.

        Returns
        -------
        dict
            server connection and status information

        """
        info = self.ep.info()
        return {
            "majorVersion": info.majorVersion,
            "build": info.build,
            "proxy": { "proxyEndpoints": [ep for ep in info.proxy.proxyEndpoints] },
        }

    def list_collections(self, starts_with=""):
        """
        Returns a list of collection paths using the `starts_with` argument for
        filtering.

        Returns
        -------
        collection paths: list[str]

        """
        return [c for some in self.ep.listCollections(starts_with) for c in some]

    def streams_in_collection(self, *collection, is_collection_prefix=True, tags=None, annotations=None):
        """
        Search for streams matching given parameters

        This function allows for searching

        Parameters
        ----------
        collection: str
            collections to use when searching for streams, case sensitive.
        is_collection_prefix: bool
            Whether the collection is a prefix.
        tags: Dict[str, str]
            The tags to identify the stream.
        annotations: Dict[str, str]
            The annotations to identify the stream.

        Returns
        ------
        list
            A list of stream objects found with the provided search arguments.

        """
        result = []

        if tags is None:
            tags = {}

        if annotations is None:
            annotations = {}

        if not collection:
            collection = [None]

        for item in collection:
            streams = self.ep.lookupStreams(item, is_collection_prefix, tags, annotations)
            for desclist in streams:
                for desc in desclist:
                    tagsanns = unpack_stream_descriptor(desc)
                    result.append(Stream(
                        self, uuidlib.UUID(bytes = desc.uuid),
                        known_to_exist=True, collection=desc.collection,
                        tags=tagsanns[0], annotations=tagsanns[1],
                        property_version=desc.propertyVersion
                    ))

        return result

    def collection_metadata(self, prefix):
        """
        Gives statistics about metadata for collections that match a
        prefix.

        Parameters
        ----------
        prefix: str
            A prefix of the collection names to look at

        Returns
        -------
        tuple
            A tuple of dictionaries containing metadata on the streams in the
            provided collection.

        """
        ep = self.ep
        tags, annotations = ep.getMetadataUsage(prefix)
        pyTags = {tag.key: tag.count for tag in tags}
        pyAnn = {ann.key: ann.count for ann in annotations}
        return pyTags, pyAnn

    def __reduce__(self):
        raise InvalidOperation("BTrDB object cannot be reduced.")
